#!/usr/bin/env python3
"""
This script creates a modelfactors table from GPU hardware counters generated 
by Rocprof.
"""

import argparse
import sys
import pandas as pd
import numpy as np
from typing import List, Tuple
from matplotlib import pyplot as plt
import seaborn as sns
from matplotlib import cm
from matplotlib.colors import Normalize

# Store hardware specifications globally.
# The Wavefronts, vgpr, and sgpr metrics contains the maximum values.
# The VGPRtoOccupancy metric is mapped as: "<=key" -> "#waves per EU".
DEVICE_SPECS = {
    "MI250X": {
        "Wavefronts": 3520,
        "vgpr": 256,
        "sgpr": 102,
        "VGPRtoOccupancy": {
            64: 8,
            72: 7,
            80: 6,
            96: 5,
            128: 4,
            168: 3,
            256: 2,
        },
    },
}

# Store metrics mapping category to hardware counters globally.
METRICS = {
    "Device Load": [
        "Wavefronts"
    ],
    "Occupancy": [
        # TODO: Add wavefronts.
        "vgpr", 
        "sgpr",
        # TODO: Add shared memory.
    ],
    "Memory Efficiency": [
        "MemUnitBusy", 
        "L2CacheHit", 
        "MemUnitStalled", 
        "LDSBankConflict",
    ],
    "Computational Efficiency": [
        "VALUUtilization", 
        "VALUBusy",
        # TODO: Add atomics.
    ],
}


def _parse_user_args() -> argparse.Namespace:
    """Parse user arguments."""
    # Setup argparser.
    parser = argparse.ArgumentParser(
                    prog="gpu_modelfactors.py",
                    description="Create GPU modelfactors from Rocprof " +
                                "GPU hardware counters csv.",
                    )
    parser.add_argument("-t", "--title", action="store_true", help="set " + \
                        "flag if you want a title added to the modelfactor " + \
                        "figures.")
    parser.add_argument("-k", "--kernel", nargs=1,
                        help="provide kernel name to process.")
    parser.add_argument("-d", "--device", nargs=1, choices=DEVICE_SPECS.keys(),
                        default="MI250X",
                        help="select specific device for percentual coloring.")
    parser.add_argument("data_paths", nargs="+", type=str,
                        help="provide paths to data CSV files with Rocprof" + \
                             "hardware counters. These paths can be " + \
                             "prepanded with the name you want to give the" + \
                             "columns, e.g.:    col1:data/file.csv")
    
    # Parse args and just print usage in case no arguments are given.
    args = parser.parse_args(args=None if sys.argv[1:] else ['--help'])

    return args


def load_data(data_paths: List[str]) -> List[Tuple[str, pd.DataFrame]]:
    """
    Parse every data path on column names, then check if the data paths exist 
    and load into Pandas DataFrames if so.
    """
    dfs = []
    
    for i, dpath in enumerate(data_paths):
        # Check if path is prepanded with column name.
        if ":" in dpath:
            col_name, path = dpath.split(":")
        else:
            col_name = f"col{i}"
            path = dpath
        
        # Check if the path points to an existing file.
        try:
            df = pd.read_csv(path)
        except FileNotFoundError:
            print(f"File at {path} not found, skipping..")
            continue
        
        dfs.append((col_name, df))

    return dfs


def prepare_df(dfs: List[pd.DataFrame], kernel: str, device: str) -> pd.DataFrame:
    """Extract only the useful data from the dataframes."""
    # Collect all hardware counters to be filtered in order of plotting.
    hwcs = [counter for v in METRICS.values() for counter in v]
    
    # Setup empty DataFrame to merge data paths.
    merged_df = pd.DataFrame(columns=hwcs)

    for col_name, df in dfs:
        # Extract requested kernel and useful columns from the DataFrames.
        new_df = df[df["KernelName"] == kernel]
        new_df = new_df[hwcs]

        # Set datatypes correctly that are inferred wrongly and add to list.
        new_df.convert_dtypes()
        new_df = new_df.astype({"Wavefronts": int, "vgpr": int, "sgpr": int})
        
        # Add first row to merged DataFrame for this input file.
        # Use the provided column name as the index.
        merged_df.loc[col_name] = new_df.iloc[0].values
        
    # Convert Wavefronts, VGPR, and SGPR to be relative to device specs.
    hwcs_device = [counter 
                   for v in [METRICS["Device Load"], METRICS["Occupancy"]] 
                   for counter in v]
        
    for metric in hwcs_device:
        merged_df[metric] = \
            merged_df[metric] / DEVICE_SPECS[device][metric] * 100

    # Return transposed version for the visualization.        
    return merged_df.transpose()


def get_df_percentages(kernel_df: pd.DataFrame, device: str) -> pd.DataFrame:
    """Get percentual changes within the kernel_df and return as DataFrame."""    
    # Create copy of kernel data to store percentages.
    perc_df = kernel_df.copy()

    # Device Load and Occupation metrics are already colored relative to device.
    # Invert scale for VGPR.
    perc_df.loc["vgpr"] = 100 - perc_df.loc["vgpr"]
    
    # Make SGPR a boolean color by using maximum as cut-off point.
    perc_df.loc["sgpr", perc_df.loc["sgpr"] <= 100] = 100
    perc_df.loc["sgpr", perc_df.loc["sgpr"] > 100] = 0

    # For the other metrics, take percentual change compared to the first 
    # column as the baseline.
    hwcs_scaling = [counter 
                   for v in [METRICS["Memory Efficiency"], 
                             METRICS["Computational Efficiency"]] 
                   for counter in v]
    perc_df.loc[hwcs_scaling, perc_df.columns[0]] = 100
    baseline_col = kernel_df.loc[hwcs_scaling, perc_df.columns[0]]

    # Transform other columns to be percentual changes of baseline.
    for col in perc_df.columns[1:]:
        perc_df.loc[hwcs_scaling, col] = 100 + 100 * \
            ((kernel_df.loc[hwcs_scaling, col] - baseline_col) / baseline_col)
    
        # Fix NaNs due to division by zero.
        if perc_df.loc[hwcs_scaling, col].isnull().values.any():
            idxs = np.where(perc_df.loc[hwcs_scaling, col].isnull().values)[0]
            
            for idx in idxs:
                # Check if it the NaN is due to a division by zero.
                # If so, simply add 0% or 100% to the dividend.
                if baseline_col.iloc[idx] == 0.0:
                    # Transform idx to row identifier.
                    metric = hwcs_scaling[idx]
                    perc_df.loc[metric, col] = 100 + kernel_df[col].iloc[idx]

    return perc_df


def plot_modelfactors(kernel_df: pd.DataFrame, kernel: str, device: str, title: bool=False) -> None:
    """Plot the modelfactors for the given kernel as a heatmap."""
    # Transform the DataFrame into one based on percentages for coloring the 
    # cells correctly.
    perc_df = get_df_percentages(kernel_df, device)
    extra_columns = max([0, len(kernel_df.columns) - 4])

    # Create figure with subplots for each category.
    # Add more horizontal space per column added.
    fig = plt.figure(figsize=(6.4 + extra_columns, 4.8))
    hratios = [len(v) for v in METRICS.values()]
    hratios.insert(2, 0)
    gs = fig.add_gridspec(len(METRICS.keys()) + 1, hspace=0.3, 
                          height_ratios=hratios)
    axs = gs.subplots(sharex=True, sharey=False)
    
    # Remove filler axis' styling and remove it from list.
    axs[2].set_axis_off()
    axs = np.delete(axs, 2)

    # Map category to subtext.
    cat_text = {
        "Device Load": "(% of device)",
        "Occupancy": "(% of device)",
        "Memory Efficiency": f"(% of {kernel_df.columns[0]})",
        "Computational Efficiency": f"(% of {kernel_df.columns[0]})",
    }
    
    # Plot the metrics in their categories.
    for ax, [cat, hwcs] in zip(axs, METRICS.items()):
        sns.heatmap(data=perc_df.loc[hwcs], annot=kernel_df.loc[hwcs], 
                    fmt=".2f", cmap="RdYlGn", linewidths=1.5, ax=ax, 
                    cbar=False, vmin=0, vmax=100)

        # Fix label and ticks.
        ax.tick_params(axis="y", labelrotation=0)
        ax.tick_params(axis="x", bottom=False, top=False, labelbottom=False)
        
        # Add category and subtext above first label.
        label_pos = ax.get_yticklabels()[0].get_position()
        ax.text(label_pos[0] - 0.1, label_pos[1] - 0.65, cat, 
                style="oblique", fontweight="heavy", ha="right", 
                fontdict={'family' : 'monospace'})
        ax.text(label_pos[0] - 0.1, label_pos[1] - 0.36, cat_text[cat], 
                style="italic", fontweight="semibold", ha="right", 
                fontsize="xx-small", fontdict={'family' : 'monospace'})

    # Only enable horizontal axis labels at the top.
    axs[0].tick_params(axis="x", top=True, labeltop=True)

    # Add colorbar to the right side of the figure.
    cmap = cm.ScalarMappable(norm=Normalize(0, 100), cmap="RdYlGn")
    fig.colorbar(cmap, label="Relative load (%)", ax=axs)

    # Add title if specified.
    if title:
        escaped_kname = kernel.replace("$", "\$")
        fig.suptitle(f"GPU Modelfactors for {escaped_kname}.", 
                     x=0.35+extra_columns*0.02, y=0.99, 
                     horizontalalignment="center")

    plt.savefig(f"figures/gpu_modelfactors_{kernel}.png", bbox_inches='tight', 
                dpi=600)
    # plt.show()


if  __name__ == "__main__":
    # Parse input arguments.
    args = _parse_user_args()

    # Load CSV data with hardware counters.
    dfs = load_data(args.data_paths)

    # Create a modelfactors for each kernel requested.
    for kernel in args.kernel:
        # Prepare dfs to create a modelfactors table.
        kernel_df = prepare_df(dfs, kernel, args.device)
        
        # Plot the heatmap with comparison.
        plot_modelfactors(kernel_df, kernel, args.device, title=args.title)
